>module Arrow where
>
>import ST
>import Monad
>class Arrow a where
>      arr :: (b -> c) -> a b c
>      (>>>) :: a b c -> a c d -> a b d -- (b |-> c) -> (c |-> d) -> (b |-> d)
>      first :: a b c -> a (b,d) (c,d)


>class Arrow a => ArrowChoice a where
>      left :: a b c -> a (Either b d) (Either c d)

>class Arrow a => ArrowApply a where
>      app :: a (a b c, b) c


>class Arrow a => ArrowZero a where
>      zeroArrow :: a b c

>class ArrowZero a => ArrowPlus a where
>      (+++) :: a b c -> a b c -> a b c

>class Arrow a => ArrowFix a where
>      fixArrow :: a (b,c) c -> a b c

>class ArrowFix a => ArrowCircuit a where
>      delay :: b -> a b b





>bind :: Arrow a => a b c -> a (b,c) d -> a b d
>u `bind` f = returnA &&& u >>> f

>bind_ :: Arrow a => a b c -> a b d -> a b d
>u `bind_` f = u `bind` (arr fst >>> f)

>returnA :: Arrow a => a b b
>returnA = arr id


>leftApp :: ArrowApply a => a b c -> a (Either b d) (Either c d)
>leftApp f = arr ((\b -> (arr (\ () -> b) >>> f >>> arr Left, ())) |||
>	     (\d -> (arr (\  () -> d) >>> arr Right, ()))) >>> app

          +--------------+
          | >>>          | 
    b --->|---<f>---<g>--|--> d
          |              |
          +--------------+

          +--------------+
          | first        | 
    b --->|---<f>--------|--> c
    d --->|--------------|--> d
          +--------------+


          +--------------+
  Either  | left         |  Either 
    b --->|---<f>--------|--> c
    d --->|--------------|--> d
          +--------------+




           +--------------+
           | app          |  Either 
 a b c --->|---<a b c>----|--> c
    b  --->|------^       |
           +--------------+


          +--------------+
          | +++          |  
    b --->|----+-<f>-+---|--> c
          |    \-<g>-/   |
          +--------------+



 recursion for arrows:

>instance Arrow (->) where
>	  arr = id
>	  (>>>) = flip (.)
>	  first f = \(x,y) -> (f x,y)

>instance ArrowChoice (->) where
>	  left f (Left b) = Left (f b)
>	  left f (Right d) = Right d

>instance ArrowApply (->) where
>	  app = \(f,x) -> f x




>fix :: Arrow a => a b b -> a b [b]
>fix a = a >>> (returnA &&& fix a) >>> arr (\ (x,y) -> (x:y) )


>second :: Arrow a => a b c -> a (d,b) (d,c)
>second f = arr swap >>> first f >>> arr swap
>	where swap (x,y) = (y,x)

          +--------------+
 Either   | <++>          |  Either 
    b --->|---<f>--------|--> c
    d --->|---------<g>--|--> e
          +--------------+

>(<++>) :: ArrowChoice a => a b c -> a d e -> a (Either b d) (Either c e)
>f <++> g = left f >>> right g

          +--------------+
  Either  | right        |  Either 
    d --->|--------------|--> d
    b --->|---<f>--------|--> c
          +--------------+

>right :: ArrowChoice a => a b c -> a (Either d b) (Either d c)
>right f = arr mirror >>> left f >>> arr mirror
>      where mirror :: Either a b -> Either b a
>	     mirror (Left x) = Right x
>	     mirror (Right x) = Left x

>instance Monad m => ArrowApply (Kleisli m) where
>	  app = K (\ ~(K f,x) -> f x)

>newtype ArrowApply a => ArrowMonad a b = M (a Int b)

>instance ArrowApply a => Monad (ArrowMonad a) where
>	  return x = M (arr (\_ -> x))
>	  M m >>= f = M (m >>>
>			 arr (\x -> let M h = f x in (h, undefined)) >>>
>			 app)

          +--------------+
 Either   | |||          |
    b --->|---<f>----+---|--> d
    c --->|---<g>---/    |
          +--------------+

>(|||) :: ArrowChoice a => a b d -> a c d -> a (Either b c) d
>f ||| g = (f <++> g) >>> arr untag
>      where untag (Left x) = x
>	     untag (Right y) = y



>newtype Kleisli m a b = K (a -> m b)

>instance Monad m => ArrowChoice (Kleisli m) where
>	  left ~(K f) = K (\x -> case x of
>			     Left b -> f b >>= \c -> return (Left c)
>			     Right d -> return (Right d))

>instance Monad m => Arrow (Kleisli m) where
>	  arr f = K (\b -> return (f b))
>	  ~(K f) >>> ~(K g) = K (\b -> f b >>= g)
>	  first ~(K f) = K (\ ~(b,d) -> f b >>= \c -> return (c,d))



       +--------------+
       | ***          |
 b --->|---<f>--------|--> c
 d --->|---------<g>--|--> e
       +--------------+

>(***) :: Arrow a => a b c -> a d e -> a (b,d) (c,e)
>f *** g = first f >>> second g

     +--------------+
  b  | &&& /--<f>---|--> c
 --->|-----+--<g>---|--> d
     +--------------+

>(&&&) :: Arrow a => a b c -> a b d -> a b (c,d)
>f &&& g = arr (\b -> (b,b)) >>> (f *** g)

>liftA2 :: Arrow a => (b -> c -> d) -> a e b -> a e c -> a e d
>liftA2 op f g = (f &&& g) >>> arr (\ ~(b,c) -> b `op` c)


>test :: Arrow a => a b Bool -> a b (Either b b)
>test f = (f &&& returnA) >>> arr (\ ~(b,x) -> if b then Left x else Right x)


>newtype MaybeFunctor a b c = MF (a b (Maybe c))

>liftMaybe :: Arrow a => a b c -> MaybeFunctor a b c
>liftMaybe f = MF (f >>> arr Just)

>instance ArrowChoice a => Arrow (MaybeFunctor a) where
>	  arr f = liftMaybe (arr f)
>	  MF f >>> MF g = MF (f >>> arr (\z -> case z of
>					       Just c -> Left c
>					       Nothing -> Right Nothing) >>>
>				(g ||| returnA))
>	  first (MF f) = MF (first f >>>
>			    arr (\ ~(c',d) -> case c' of
>					     Just c -> Just (c,d)
>					     Nothing -> Nothing))

>instance ArrowChoice a => ArrowZero (MaybeFunctor a) where
>	  zeroArrow = MF (arr (\_ -> Nothing))

>instance ArrowChoice a => ArrowPlus (MaybeFunctor a) where
>	  MF f +++ MF g = MF ( (f &&& returnA) >>>
>			       arr (\ ~(c',b) -> case c' of
>						Just _ -> Left c'
>						Nothing -> Right b) >>>
>				(returnA ||| g))

>instance ArrowChoice a => ArrowChoice (MaybeFunctor a) where
>	  left (MF f) = MF (left f >>> arr (\x -> case x of
>			Left (Just c) -> Just (Left c)
>			Left Nothing  -> Nothing
>			Right d	      -> Just (Right d)))

>instance (ArrowChoice a, ArrowApply a) => ArrowApply (MaybeFunctor a) where
>	  app = MF (arr (\ ~(MF f,b) -> (f,b)) >>> app)


STArrow is based on the state functor from John Hughes' "Generalising
monads to Arrows" paper.

>type STArrow s b c = Kleisli (ST s) b c

>newtype StateFunctor s a b c = SF (a (b,s) (c,s))

>liftState :: Arrow a => a b c -> StateFunctor s a b c
>liftState f = SF (first f)


>class StateAspect sa s where
>      fetch :: sa b s
>      store :: sa s ()

>instance Arrow a => StateAspect (StateFunctor s a) s where
>	  fetch = SF (arr (\ (_,s) -> (s,s)))
>	  store = SF (arr (\ (x,s) -> ((),x)))


>instance Arrow a => Arrow (StateFunctor s a) where
>       arr f = liftState (arr f)
>       SF f >>> SF g = SF (f >>> g)
>       first (SF f) = SF (arr (\ ((b,d),s) -> ((b,s),d)) >>> first f
>                         >>> arr (\ ((c,s),d) -> ((c,d),s)))


fetch :: Arrow a => StateFunctor s a b s
fetch = SF (arr (\ (_,s) -> (s,s)))

store :: Arrow a => StateFunctor s a s ()
store = SF (arr (\ (x,s) -> ((),x)))

>instance ArrowChoice a => ArrowChoice (StateFunctor s a) where
>	  left (SF f) = SF ( arr (\ (x,s) -> case x of
>                              Left b -> Left (b,s)
>			       Right c -> Right (c,s)) >>>
>			((f >>> first (arr Left)) ||| first (arr Right)))

>instance ArrowZero a => ArrowZero (StateFunctor s a) where
>	  zeroArrow = SF zeroArrow

>instance ArrowPlus a => ArrowPlus (StateFunctor s a) where
>	  SF f +++ SF g = SF (f +++ g)

>instance ArrowApply a => ArrowApply (StateFunctor s a) where
>	  app = SF (arr (\ ((SF f,b),s) -> (f,(b,s))) >>> app)

Continuation passing:

>newtype CPSFunctor ans a b c = CPS ((a c ans) -> (a b ans))

>liftCPS :: Arrow a => a b c -> CPSFunctor ans a b c
>liftCPS f = CPS (\k -> f >>> k)

>instance ArrowApply a => Arrow (CPSFunctor ans a) where
>	  arr f = liftCPS (arr f)
>	  CPS f >>> CPS g = CPS (f . g)
>	  first (CPS f) = CPS (\k -> arr (\ (b,d) -> 
>		(f (arr (\c -> (c,d)) >>> k),b)) >>> app)

>jump :: ArrowApply a => CPSFunctor ans a (a c ans,c) x
>jump = CPS (\_ -> app)

>callcc :: ArrowApply a => (a c ans -> CPSFunctor ans a b c) 
>	                    -> CPSFunctor ans a b c
>callcc f = CPS (\k -> let CPS g = f k in g k)


