>module Main where

import FastParser

>import CommonParser
>import DocParser (pdocument, tokenToTree,Token)
>import Language
>import Parser (LineInfo,emptyLineInfo)
>import Specification
>import System (exitWith, ExitCode(..))
>import WhiteElim(whiteelim)
>import RedElim(redelim_forest)
>import Util
>import UsageElim(usageelim_forest)
>import ColorNesting(colornesting_forest)
>import Timer
>import System (getArgs)
>import Monad (foldM)

>main :: IO ()
>main = do args <- getArgs
>	   maxtime <- case args of
>		[] -> return 5000
>		[s] -> return (read s)
>		x -> fail "Error: Too many command line arguments. The only argument is the time allowed for the processing (in seconds)."
>	   input <- getContents
>	   result <- start_processing maxtime input (process input)
>	   putStr result

>process :: String -> TimerHandle String -> IO String
>process input handle = do
>	   -- debugM_ "input" $ return input
>	   let parser :: CommonParser Char LineInfo [Token]
>	       parser = pdocument 
>	   pres' <- parseP parser emptyLineInfo input
>	   -- debugM_ ("parsed",pres) $ return pres
>	   parserresult <- case pres' of
>		ParseFail m _ -> do putStrLn $ render m
>				    exitWith (ExitFailure 1)
>		ParseOk v [] _ -> return v
>		ParseOk _ re  _ -> do putStrLn ("Parse failure, extra input:"
>					    ++ (show (take 5 re)))
>				      exitWith (ExitFailure 1)
>	   pres <- tokenToTree [] parserresult [] >>= (return . In . DSeq)
>	   let tags = parser_to_tags pres
>	   let m = meaning rootcontext tags
>	   let ctree = map (buildrecctree rootcontext) tags
>	   let produce_output optimized =
>	         concat (map show (map output optimized))
>	   let run_optim inp optim = do
>			 let optimized = optim rootcontext inp
>			 send_result handle $! (produce_output optimized)
>			 return optimized
>	   all_optims_result <- foldM run_optim ctree optimization_steps

          let semantics1 = m
	       semantics2 = meaningofctree all_optims_result
	       comp = to_string (semantics1 == semantics2) 
	       filtered = filter (\ (x,y) -> x /= y) (zip semantics1 semantics2)
	       duplicate = concatMap (\ (x,y) -> [ch x,ch y]) (zip semantics1 semantics2)
              to_string True = "OK"
              to_string False = "FAIL" 
		in debugM_ ("Correctness: ", comp) $ return comp

>            -- debugM_ ("Correctness: ",(comp,zip semantics1 semantics2),filtered,enclose duplicate) $ return comp


>	   return (produce_output all_optims_result)



>output :: Rec CTree -> Rec TagTree
>output tree = fold selectnode tree
>   where selectnode :: CTree (Rec TagTree) -> Rec TagTree
>         selectnode (CT (ContextEntry a _)) = In a

>optimization_steps :: [Context -> [Rec CTree] -> [Rec CTree]]
>optimization_steps = [
>                     \_ t -> map whiteelim t,
>                      \_->(colornesting_forest []),
>                     redelim_forest, 
>		      \c -> map (replaceContext c) . usageelim_forest,
>                     \_ t -> map whiteelim t,
>                      \_->(colornesting_forest []), 
>                     redelim_forest, 
>		      \c -> map (replaceContext c) . usageelim_forest
>                     ]


