>module RedElim (redelim, redelim_forest) where
>import Util
>import Specification

Redundancy elimination will first try to turn trees of the form

    A
   / \
  B   C
  |   |\
  D   E F

into
    A__
  /  \ \
  B   E F
  |
  D

based on whether the A and C nodes have the same semantics. This
will always reduce the size of the output. Note that Both A and C nodes
will have to be TagNodes for this to do anything.

There are two entry points into this, "redelim" works for trees
and "redelim_forest" works for forests of trees (and it takes
the context where the forest is to be interpreted as argument).

>redelim :: Rec CTree -> Rec CTree
>redelim = fold elim_redundancy_fold

>redelim_forest :: Context -> [Rec CTree] -> [Rec CTree]
>redelim_forest c lst = shallow_redelim_forest c (map redelim lst)

>shallow_redelim_forest :: Context -> [Rec CTree] -> [Rec CTree]
>shallow_redelim_forest c lst = let 
>		elimBool = map willElim lst
>		willElim (In (CT (ContextEntry (TagNode t1 _) c'))) = 
>			 redundancy_eq t1 c c'
>		willElim _ = False
>		transElimed (True, 
>		      In (CT (ContextEntry (TagNode _ lst2) _))) = lst2
>		transElimed (True, x) = [x]
>		transElimed (False,x) = [x]
>              in concatMap transElimed (zip elimBool lst)

>elim_redundancy_fold :: CTree (Rec CTree) -> Rec CTree
>elim_redundancy_fold (CT (ContextEntry (TagNode t lst) c)) =
>	In $ CT (ContextEntry (TagNode t (shallow_redelim_forest c lst)) c)
>elim_redundancy_fold x = In x

                         b  em i s tt u size color

>redundancy_eq :: Tag -> Context -> Context -> Bool
>redundancy_eq BTag  (PRec b _ _ _ _ _ _ _) (PRec b' _ _ _  _ _ _ _) = b == b'
>redundancy_eq EMTag (PRec _ _ _ s _ _ _ _) (PRec _  _ _ s' _ _ _ _) = s && s'
>redundancy_eq ITag  (PRec _ _ i _ _ _ _ _) (PRec _  _ i' _ _ _ _ _) = i == i'
>redundancy_eq PLTag (PRec b e i s t u _ _) (PRec b' e' i' s' t' u' _ _) =
>			   b == b' && e == e' && i == i' && s == s'
>			   && t == t' && u == u'
>redundancy_eq STag  (PRec _ e _ s _ _ _ _) (PRec _ e' _ s' _ _ _ _) =
>	       e == e' && s == s'
>redundancy_eq TTTag (PRec _ _ _ _ t _ _ _) (PRec _ _ _ _ t' _ _ _) = t == t'
>redundancy_eq UTag  (PRec _ _ _ _ _ u _ _) (PRec _ _ _ _ _ u' _ _) = u == u'
>redundancy_eq (ColorTag c) (PRec _ _ _ _ _ _ _ c1) (PRec _ _ _ _ _ _ _ c2) =
>	       c1 == c2 && c2 == c
>redundancy_eq (SizeTag s) (PRec _ _ _ _ _ _ s1 _) (PRec _ _ _ _ _ _ s2 _) =
>	       s1 == s2 && Sz s == s2
