>module Specification where
>import Util
>import Language
>import Maybe(fromJust)
>import DocParser(isSpaceChar)
>import PrettyP

>meaning :: Context -> [Rec TagTree] -> [DecoratedCharacterType]
>meaning c = meaningofctree . map (buildrecctree c)

>meaningofctree :: [Rec CTree] -> [DecoratedCharacterType]
>meaningofctree = compresswhitespace Nothing . concatMap recmeaning

>data ColorType = Red |Green |Blue | Cyan |Magenta |Yellow |Black | White
>		| Transparent
>        deriving(Eq)

>instance Show ColorType where
>	  showsPrec _ x = showString $ fromJust $ lookup x table
>		where table = [(Red,"r"),(Green,"g"),(Blue,"b"),
>				(Cyan,"c"),(Magenta,"m"),
>				(Yellow,"y"), (Black,"k"),
>				(White,"w"), (Transparent,"_")]

>data SizeType = UnknownSz | Sz Integer deriving (Eq)
>data DecoratedCharacterType = DCT { ch :: Char,  context :: PropertyRecordType }
>    deriving(Eq)

>data Tag = BTag | EMTag | ITag | PLTag | STag | TTTag | UTag
>  | SizeTag Integer | ColorTag ColorType
>        deriving(Eq)

>data TagTree a = TagNode Tag [a]
>		    | CharNode [Char]
>		    | SpaceNode [Char]

>data TagNodeType = TagType |CharType |SpaceType
>    deriving (Eq)

>data ContextTree a = ContextEntry a Context

>newtype CTree a = CT (ContextTree (TagTree a))
>type Context = PropertyRecordType

>data PropertyRecordType = PRec { b_prop :: Bool,
>				   em_prop :: Bool,
>				   i_prop :: Bool,
>				   s_prop :: Bool,
>				   tt_prop :: Bool,
>				   u_prop :: Integer,
>				   size_prop :: SizeType,
>				   color_prop :: ColorType }
>   deriving(Eq)

>class TreeNodeBase t where
>	tag :: t -> Tag   -- requires: has tag
>	nodecontext :: t -> Context
>	chars :: t -> String -- requires: has chars
>	nodetype :: t -> TagNodeType

>class TreeNodeBase t => TreeNode t a where
>	child :: t -> [a]
>	tagtree :: t -> TagTree a
>	setchilds :: [a] -> t -> t  -- requires: has childs.


>instance Show SizeType where
>	  showsPrec _ (Sz i) = shows i
>	  showsPrec _ UnknownSz = showChar '_'

>instance Show PropertyRecordType where
>	showsPrec _ (PRec b em i s tt u size color) = ifp b 'b' . ifp em "em" .
>			ifp i 'i' . ifp s 's' . ifp tt "tt" .
>			shows u . shows size . shows color
>		where ifp x c = if x then shows c else showChar '.'

>rootcontext :: PropertyRecordType
>rootcontext = PRec False False False False False 0 UnknownSz Transparent

>compareDecoratedCharTypes :: [DecoratedCharacterType] -> [DecoratedCharacterType] -> Bool
>compareDecoratedCharTypes (x1:xs1) (x2:xs2) = x1 == x2 && compareDecoratedCharTypes xs1 xs2
>compareDecoratedCharTypes [] (_:_) = False
>compareDecoratedCharTypes (_:_) [] = False
>compareDecoratedCharTypes [] [] = True



>instance PpShow DecoratedCharacterType where
>	  pp x = pp (show x)

>instance Show DecoratedCharacterType where
>	showsPrec _ (DCT c ctx) = showChar c . showString "::" . shows ctx 

.
		    showChar '\n'

>instance Show Tag where
>	showsPrec _ BTag = showChar 'B'
>	showsPrec _ EMTag = showString "EM"
>	showsPrec _ ITag = showChar 'I'
>	showsPrec _ STag = showChar 'S'
>	showsPrec _ TTTag = showString "TT"
>	showsPrec _ UTag = showChar 'U'
>	showsPrec _ PLTag = showString "PL"
>	showsPrec _ (SizeTag i) = shows i
>	showsPrec _ (ColorTag c) = shows c

>instance PpShow Tag where
>	pp t = pp (show t)

>instance Show a => Show (TagTree a) where
>	showsPrec _ (TagNode t lst) = showChar '<' . shows t . showChar '>' .
>		   foldr (.) id (map shows lst)
>			. showString "</" . shows t . showChar '>'
>	showsPrec _ (CharNode lst) = showString lst
>	showsPrec _ (SpaceNode lst) = showString lst

>allowedTags1 :: [Tag]
>allowedTags1 = [BTag, EMTag, ITag, PLTag, STag, TTTag, UTag, SizeTag 0, SizeTag 1, SizeTag 2, SizeTag 3, SizeTag 4, SizeTag 5, SizeTag 6, SizeTag 7, SizeTag 8, SizeTag 9, ColorTag Red, ColorTag Green, ColorTag Blue, ColorTag Cyan, ColorTag Magenta, ColorTag Yellow, ColorTag Black, ColorTag White] 
>converttag :: DocTag -> Tag
>converttag s = fromJust $ lookup s (zip allowedTags allowedTags1)

>parser_to_tags :: Rec Document -> [Rec TagTree ] 
>parser_to_tags (In (DSeq lst)) = concatMap parser_to_tags lst
>parser_to_tags (In (DTag dt x)) = [In $TagNode (converttag dt) (parser_to_tags x)]
>parser_to_tags (In (DText chs)) = [In $CharNode chs]
>parser_to_tags (In (DSpace chs)) = [In $SpaceNode chs]

>instance Functor TagTree where
>	fmap f (TagNode t lst) = TagNode t (fmap f lst)
>	fmap _ (CharNode c) = CharNode c
>	fmap _ (SpaceNode c) = SpaceNode c


>instance Show a => Show (ContextTree (TagTree a)) where
>	showsPrec _ (ContextEntry x@(TagNode _ _) _) = shows x
>	showsPrec _ (ContextEntry x y) = shows (x,y)

>instance Functor ContextTree where
> 	fmap f (ContextEntry x y) = ContextEntry (f x) y



>instance TreeNodeBase (CTree a) where
>	tag (CT (ContextEntry (TagNode t _) _)) = t
>	tag _ = error "Tag used for nontag"
>	chars (CT (ContextEntry (CharNode chs) _)) = chs
>	chars (CT (ContextEntry (SpaceNode chs) _)) = chs
>	chars _ = error "chars with non-char/non-space node"
>	nodetype (CT (ContextEntry (TagNode _ _) _)) = TagType
>	nodetype (CT (ContextEntry (CharNode _) _)) = CharType
>	nodetype (CT (ContextEntry (SpaceNode _) _)) = SpaceType
>	nodecontext (CT (ContextEntry _ c)) = c

>instance TreeNode (CTree a) a where
>	child (CT (ContextEntry (TagNode _ lst) _)) = lst
>	child _ = []
>	tagtree (CT (ContextEntry t _)) = t
>	setchilds lst (CT (ContextEntry (TagNode t _) c)) =
>		CT $ ContextEntry (TagNode t lst) c
>	setchilds _ _ = error "No childs on char or space nodes"

>instance TreeNodeBase (TagTree a) where
>	tag (TagNode t _) = t
>	tag _ = error "Tag used for nontag"
>	nodecontext _ = error "No context in TagTree"
>	chars (CharNode chs) = chs
>	chars _ = error "charnodes from non-char/non-space types"
>	nodetype (TagNode _ _)  = TagType
>	nodetype (CharNode _)  = CharType
>	nodetype (SpaceNode _) = SpaceType

>instance TreeNode (TagTree a) a where
>	child (TagNode _ lst) = lst
>	child _ = []
>	tagtree t = t
>	setchilds lst (TagNode t _) = TagNode t lst
>       setchilds _ _ = error "No childs on char or space nodes" 


>instance TreeNodeBase (Rec CTree) where
>	tag (In (CT (ContextEntry (TagNode t _) _))) = t
>	tag _ = error "Tag used for non-tag"
>	nodecontext (In (CT (ContextEntry _ c))) = c
>	chars (In (CT (ContextEntry (CharNode chs) _))) = chs
>	chars (In (CT (ContextEntry (SpaceNode chs) _))) = chs
>	chars _ = error "chars from non char/space nodes"
>	nodetype (In (CT (ContextEntry (TagNode _ _) _))) = TagType
>	nodetype (In (CT (ContextEntry (CharNode _) _))) = CharType
>	nodetype (In (CT (ContextEntry (SpaceNode _) _))) = SpaceType

>instance TreeNode (Rec CTree) (Rec CTree) where
>	child (In (CT (ContextEntry (TagNode _ lst) _))) = lst
>	child _ = []
>	tagtree (In (CT (ContextEntry t _))) = t
>	setchilds lst (In (CT (ContextEntry (TagNode t _) c))) = In $CT $ ContextEntry (TagNode t lst) c
>       setchilds _ _ = error "No childs on char or space nodes" 



>instance Show a => Show (CTree a) where
>	showsPrec _ (CT x) = shows x

>instance Functor CTree where
>	fmap f (CT (ContextEntry x c)) = CT (ContextEntry (fmap f x) c)


Design decision: Contexts in tagnodes include
the attributes modified by the node.

>buildrecctree :: Context -> Rec TagTree -> Rec CTree
>buildrecctree c (In (TagNode t lst)) = In $ CT $ ContextEntry 
>		(TagNode t (map (buildrecctree newc) lst))
>		newc
>     where newc = contextchange t c
>buildrecctree c (In (CharNode lst)) = In $ CT $ ContextEntry
>		(CharNode lst) c
>buildrecctree c (In (SpaceNode lst)) = In $ CT $ContextEntry
> 		(SpaceNode lst) (spacecontext c)

>spacecontext :: Context -> Context
>spacecontext c = c { s_prop = False, 
>			em_prop = False,
>			i_prop = False,
>			b_prop = False,
>			color_prop = w_if_u_unset c (u_prop c) }

>w_if_u_unset :: Context -> Integer -> ColorType
>w_if_u_unset _ 0 = White
>w_if_u_unset c _ = color_prop c

>contextchange :: Tag -> Context -> Context
>contextchange BTag c = c { b_prop = True }
>contextchange ITag c = c { i_prop = True }
>contextchange PLTag c = c { u_prop = 0, s_prop = False, em_prop = False, i_prop = False, b_prop = False, tt_prop = False }
>contextchange STag c = c {s_prop = True, em_prop = False }
>contextchange TTTag c = c { tt_prop = True }
>contextchange UTag c |u_prop c < 3 = c {u_prop = (u_prop c)+1 }
>contextchange UTag c = c
>contextchange (SizeTag i) c = c {size_prop = Sz i }
>contextchange (ColorTag color) c = c {color_prop = color }
>contextchange EMTag c | s_prop c = c
>contextchange EMTag c = c { em_prop = not (em_prop c) }

>recmeaning :: Rec CTree -> [DecoratedCharacterType]
>recmeaning tree = fold attachcontext tree

>iswhitespace :: DecoratedCharacterType -> Bool
>iswhitespace c = isSpaceChar (ch c)
 
>eqsctx :: Maybe Context -> Maybe Context -> Bool
>eqsctx _ Nothing = False
>eqsctx Nothing _ = False
>eqsctx (Just c1) (Just c2) = tt_prop c1 == tt_prop c2 &&
>	u_prop c1 == u_prop c2 &&
>	size_prop c1 == size_prop c2 &&
>	color_prop c1 == color_prop c2

>compresswhitespace,compresswhitespace' :: 
>	Maybe Context -> [DecoratedCharacterType] -> [DecoratedCharacterType]

compresswhitespace c x = debug ("compresswhitespace",show c,x) $ compresswhitespace' c x

>compresswhitespace = compresswhitespace'

>compresswhitespace' c (ch1 : sx) 
>     |iswhitespace ch1 && tt_prop ctx = 
>		     (DCT (ch ch1) curSpaceContext) 
>		     : compresswhitespace (Just curSpaceContext) sx
>     | iswhitespace ch1 && (Just curSpaceContext) `eqsctx` c = 
>		     compresswhitespace c sx
>     | iswhitespace ch1 = DCT (ch ch1) curSpaceContext
>		     : compresswhitespace (Just curSpaceContext) sx
>     | otherwise = ((DCT (ch ch1) ctx)
>		     : compresswhitespace Nothing sx)
>   where curSpaceContext = (spacecontext ctx)
>	  ctx = context ch1
>compresswhitespace' _ []= []

>attachcontext :: CTree [DecoratedCharacterType] -> [DecoratedCharacterType]
>attachcontext (CT (ContextEntry (TagNode _ a) _)) = 
>			concatMap (map characterchangefunc) a
>          where characterchangefunc x = DCT (ch x) (context x)

>attachcontext (CT (ContextEntry (CharNode x) c)) = map (`DCT` c) x
>attachcontext (CT (ContextEntry (SpaceNode x) c)) = map (`DCT` c) x


Divide & Conquer Support for Rec CTree

>indivisible :: Rec CTree -> Bool
>indivisible (In (CT (ContextEntry (TagNode _ _) _))) = False
>indivisible (In _) = True

>divide :: Rec CTree -> [Rec CTree]
>divide (In (CT (ContextEntry (TagNode _ ar) _))) = ar
>divide (In _) = []


>stripContext :: Rec CTree -> Rec TagTree
>stripContext rctree = fold stripCT rctree
>   where stripCT :: CTree (Rec TagTree) -> Rec TagTree
>	  stripCT ct = In (tagtree ct)

>replaceContext :: Context -> Rec CTree -> Rec CTree
>replaceContext c = buildrecctree c . stripContext 





